# -*- coding: utf-8 -*-

# Netsphere Financial Front / ネットスフィア経理フロント
# Copyright (C) 2007-2009,2011 HORIKAWA Hisashi. All rights reserved.
#     http://www.nslabs.jp/bside/
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


# 経理だけが使う関係
class KeiriController < ApplicationController
  before_filter :login_required
  before_filter :role_keiri_required
  

  def index
  end

  # 仕訳確認画面
  # TODO: 経理承認したときに、出力待ちテーブルにpushしたほうがいいかなぁ
  def export
    @company = Company.find(params[:company] || 1)
    functional_currency = @company.functional_currency
    raise "functional currency undefined." if !functional_currency
    fixed_day = @company.fixed_gl_day

    @entries = []
    entry_ids = []; si_entry_ids = []
    @deferred = []
    @deferred_si = []
    global = GlobalSetting.find(:first)

    # 購買請求書
    invoices = Invoice.find :all,
                    :conditions => ['state = 10 AND company_setting_id = ?',
                                    @company.id],
                    :order => 'buy_day'
    invoices.each {|inv|
      entry_day = [inv.buy_day, fixed_day + 1].max
      if inv.currency_id != functional_currency.id &&
                  entry_day > global.currency_limit_day
        @deferred << inv.id
      else
        @entries += inv.make_entries(@company, entry_day)
        entry_ids << inv.id
      end
    }

    # 販売請求書
    invoices = SalesInvoice.find :all,
                     :conditions => ['state = 10 AND company_setting_id = ?',
                                     @company.id],
                     :order => 'sell_day'
    invoices.each {|inv|
      entry_day = [inv.sell_day, fixed_day + 1].max
      if inv.currency_id != functional_currency.id &&
                          entry_day > global.currency_limit_day
        @deferred_si << inv.id
      else
        @entries += inv.make_entries(@company, entry_day)
        si_entry_ids << inv.id
      end
    }
    
    # 入金伝票
    receiving_entry_ids = []
    receivings = ReceivingSlip.find :all,
                    :joins => 'JOIN our_banks ON receiving_slips.our_bank_id = our_banks.id',
                    :conditions => ['state = 10 AND our_banks.company_setting_id = ?',
                                    @company.id],
                    :order => "date"
    receivings.each {|r|
      next if r.receivings.length == 0
      @entries += r.make_entries(@company, r.date) # 伝票日付は入金日
      receiving_entry_ids << r.id
    }

    # 出金伝票
    payment_ids = []
    payments = PaymentSlip.find :all,
                 :conditions => ["state = 10 AND company_id = ?", @company.id],
                 :order => "date"
    payments.each do |r|
      next if r.payments.length == 0 || r.payment_banks.length == 0
      @entries += r.make_entries(@company, r.date)
      payment_ids << r.id
    end

    @size = entry_ids.size + si_entry_ids.size + receiving_entry_ids.size + payment_ids.size
    session[:entries] = entry_ids
    session[:si_entries] = si_entry_ids
    session[:receiving_entries] = receiving_entry_ids
    session[:payment_entries] = payment_ids
  end


  # 実際の出力
  def export2
    company = Company.find params[:company]
    fixed_day = company.fixed_gl_day
    
    @entries = []
    
    # 購買請求書
    inv_ids = session[:entries]
    inv_ids.each {|inv_id|
      inv = Invoice.find inv_id
      entry_day = [inv.buy_day, fixed_day + 1].max
      @entries += inv.make_entries(company, entry_day)
    }
    
    # 販売請求書
    si_ids = session[:si_entries]
    si_ids.each {|inv_id|
      inv = SalesInvoice.find inv_id
      entry_day = [inv.sell_day, fixed_day + 1].max
      @entries += inv.make_entries(company, entry_day)
    }
    
    # 入金伝票
    receiving_ids = session[:receiving_entries]
    receiving_ids.each {|r_id|
      r = ReceivingSlip.find r_id
      entry_day = r.date
      @entries += r.make_entries(company, entry_day)
    }

    # 出金伝票
    payment_ids = session[:payment_entries]
    payment_ids.each do |p_id|
      r = PaymentSlip.find p_id
      entry_day = r.date
      @entries += r.make_entries(company, entry_day)
    end

    headers["Content-Type"] = "application/vnd.ms-excel; charset=UTF-8"
    headers["Content-Disposition"] = "attachment;\n  filename*=iso-2022-jp'ja'" +
                                     NKF.nkf("-m0 -j -W", "purchase.csv")
    render :action => 'export2', :layout => false

    # 連動済みとしてマークする。
    export_at = Time.now
    Invoice.transaction {
      inv_ids.each {|inv_id|
        inv = Invoice.find inv_id
        inv.state = 20
        inv.journal_export_at = export_at
        inv.journal_day = [inv.buy_day, fixed_day + 1].max
        inv.save!
      }
      
      si_ids.each {|inv_id|
        inv = SalesInvoice.find inv_id
        inv.state = 20
        inv.journal_export_at = export_at
        inv.journal_day = [inv.sell_day, fixed_day + 1].max
        inv.save!
      }
      
      # 入金伝票
      receiving_ids.each {|r_id|
        r = ReceivingSlip.find r_id
        r.state = 20
        r.journal_export_at = export_at
        r.save!
      }

      payment_ids.each do |p_id|
        r = PaymentSlip.find p_id
        r.state = 20
        r.journal_export_at = export_at
        r.save!
      end
    } # of transaction
  end


  # 総勘定元帳 (仕訳) のインポート画面
  def import_gl
    require 'csv'

    @fname = '/shiwake.csv'
    if FileTest.exist?(IMPORT_DATA_DIR + @fname)
      data = sjis2utf File.read(IMPORT_DATA_DIR + @fname)
      @head = []
      
      idx = 0
      CSV::Reader.parse(data) {|line|
        idx += 1; next if idx == 1
        break if idx == 10
        @head << line
      }
    end
  end

  
  SCRIPT_DIR = File.dirname(__FILE__) + '/../../script'
  def do_import_gl
    company = Company.find params[:import][:company_id]

    `#{KeiriController::SCRIPT_DIR}/run_import_gl.rb #{company.id}`
    redirect_to :action => 'import_gl_log', :company => company
  end

  
  # ログを表示
  def import_gl_log
    @out = File.read KeiriController::SCRIPT_DIR + '/import_gl_out.log'
    @err = File.read KeiriController::SCRIPT_DIR + '/import_gl_err.log'
  end

  
  # 取引先（支払先）のインポート画面
  def import_partners
    `#{KeiriController::SCRIPT_DIR}/run_import_partners.rb`
    redirect_to :action => 'import_partners_log'
  end
  
  # 勘定科目のインポート
  def import_account_titles
    `#{KeiriController::SCRIPT_DIR}/run_import_account_titles.rb`
    redirect_to :action => 'import_account_titles_log'
  end
  
  # 社員マスタアップロード画面
  def import_workers
  end

  # ログ表示
  def import_partners_log
    @out = File.read KeiriController::SCRIPT_DIR + '/import_partners_out.log'
    @err = File.read KeiriController::SCRIPT_DIR + '/import_partners_err.log'
  end
  
  # 勘定科目のインポートのログ表示
  def import_account_titles_log
    @out = File.read KeiriController::SCRIPT_DIR + '/import_account_titles_out.log'
    @err = File.read KeiriController::SCRIPT_DIR + '/import_account_titles_err.log'
  end

end
