# -*- coding:utf-8 -*-

# Netsphere Financial Front / ネットスフィア経理フロント
# Copyright (C) 2007-2009,2011 HORIKAWA Hisashi. All rights reserved.
#     http://www.nslabs.jp/bside/
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


# 売上側の請求書 (売掛金管理)
class SalesInvoicesController < ApplicationController
  before_filter :login_required
  before_filter :role_keiri_required, :only => [:sign_off_list, :sign_off]
  before_filter :role_apar_required

  # GETs should be safe (see http://www.w3.org/2001/tag/doc/whenToUseGet.html)
  verify :method => :post, :only => [ :destroy, :create, :update, :sign_off ],
         :redirect_to => { :action => :index }


  # 汎用の一覧ページ
  def index
    @sales_invoices = SalesInvoice.paginate :page => params[:page],
                                            :per_page => 20
  end
  

  # XSL-FOをPDFに変換する
  def render_pdf tmpl_name
    require 'tempfile'

    raise "internal error. FOP_HOME missing" if !ENV["FOP_HOME"]

    inf = Tempfile.new('body', RAILS_ROOT + "/tmp")
    inf.write render_to_string(:action => tmpl_name, :layout => false)
    inf.close

    (outf = Tempfile.new('out', RAILS_ROOT + "/tmp")).close
    (errf = Tempfile.new('err', RAILS_ROOT + "/tmp")).close

    if @company.logo_image
      File.open(RAILS_ROOT + "/tmp/logo.png", "wb") do |fp|
        fp.write @company.logo_image
      end
    else
      begin File.unlink RAILS_ROOT + "/tmp/logo.png"; rescue; end
    end

    # headers['Content-Type'] = 'application/pdf'
    r = system "#{ENV['FOP_HOME']}/fop -c #{RAILS_ROOT}/config/fop-conf.xml #{inf.path} -pdf #{outf.path} 2>#{errf.path}"
    
    if !r
      @xml_fname = inf.path
      @err_fname = errf.path
      raise "internal error"
    else
      send_file outf.path, :type => "application/pdf", :disposition => "inline"
    end

    inf.unlink
    outf.unlink
    errf.unlink
    begin File.unlink RAILS_ROOT + "/tmp/logo.png"; rescue; end
  end
  private :render_pdf
  

  def update_si_number
      to_save = false
      if !@invoice.si_number
        @invoice.si_number = (@company.si_counter += 1)
        @company.save!
        to_save = true
      end
      if !@invoice.issued_by
        @invoice.issued_by = current_user
        to_save = true
      end

      @invoice.save! if to_save
  end
  private :update_si_number


  # 作った請求書を出力
  def print_invoice
    @invoice = SalesInvoice.find params[:id]
    lang = params[:lang]
    raise SecutiryError if lang !~ /^[a-z][a-z]$/

    @company = @invoice.company_setting
    @contact = OurContact.find :first,
                :conditions => ["lang = ? AND company_setting_id = ?",
                                lang, @company.id] #"
    raise "contact not found" if !@contact

    SalesInvoice.transaction do
      update_si_number
      render_pdf "_print_invoice_" + lang

      # ログを記録しておく (表示するかは別)
      il = SiLog.new :sales_invoice => @invoice,
                     :kind => SiLog::PRINT_INVOICE,
                     :done_by => current_user,
                     :remarks => "",
                     :done_at => Time.now
      il.save!
    end
#    rescue
#      render :action => "render_pdf_error"
#    end
  end
 

  def show
    @invoice = SalesInvoice.find(params[:id])
  end
  
  # 新規作成画面
  def new
    if !current_user.default_division
      flash[:alert] = "current_user does not have any division." 
      redirect_to :controller => "users", :action => "show", :id => current_user
      return
    end

    @partner = Partner.find params[:partner]
    @company = Company.find params[:company]

    @invoice = SalesInvoice.new(
      :company_setting => @company,
      :contact_name1 => @partner.name
    )
  end

  # 明細行のレコードを作る
  def create_details new_exp
    result = []
    # raise params.inspect # DEBUG
    (0..9).each {|i| # magic
      pa = params["detail#{i}"]
      pa[:vat_code] = pa[:vat_code].to_i != 0 ? 20 : 0

      detail = SiDetail.new pa
      detail.attributes = {
        :sales_invoice => @invoice,
        :unit_price => (pa[:unit_price].blank? ? nil :
                                  nominal_amount(pa[:unit_price], new_exp) ),
        # :vat_code => pa[:vat_code].to_i != 0 ? 20 : 0,
        :line_no => i + 1
      }
      if detail.unit_price.to_i != 0 || detail.qty.to_i != 0
        detail.amount = detail.unit_price * detail.qty
      else
        detail.amount = nominal_amount(pa[:amount], new_exp)
      end
      result << detail
    }
    # 末尾の空白行を取り除く
    while last = result.last
      if last.amount != 0 || last.description.strip != ''
        break
      else
        result.pop
      end
    end
    return result
  end
  private :create_details


  def journals_from_formdata
    lines = params[:lines].to_i
    raise "internal error" if lines == 0

    journals = []
    sum_amount = 0
    (0 .. lines - 1).each {|i|
      if (ja = params["journal#{i}"])[:amount].to_f != 0.0
        journal = SiJournal.new ja
        journal.attributes = {
          :sales_invoice => @invoice,
          :amount => nominal_amount(ja[:amount], @invoice.currency.exponent),
        }
        journals << journal
        sum_amount += journal.amount
      end
    }
    
    if journals.empty?
      # 1行は作る
      journal = SiJournal.new params["journal0"]
      journal.attributes = {
        :sales_invoice => @invoice,
        :amount => 0
      }
      journals << journal
    end

    return [journals, sum_amount]
  end
  private :journals_from_formdata

  # create() / update() から呼び出される
  def create_from_formdata partner, orig_inv
    if !orig_inv
      invoice = SalesInvoice.new params[:invoice]
    else
      invoice = orig_inv
      invoice.attributes = params[:invoice]
    end
    
    invoice.si_details.clear
    details = create_details(invoice.currency.exponent)
    details.each {|r|
      invoice.si_details << r
    }
    
=begin
    invoice.si_journals.clear
    journals = create_journals invoice, params[:journal], details
    journals.each {|r|
      invoice.si_journals << r
    }
=end
    invoice.our_banks.clear
    banks = params[:ba].blank? ? [] : OurBank.find(params[:ba])
    banks.each {|b|
      invoice.our_banks << b
    }

    if !orig_inv
      invoice.attributes = {
        :partner => partner,
        :division => current_user.default_division,
        :created_by => current_user,
        :state => 0,
      }
    end

    return invoice
  end
  private :create_from_formdata


  # 作成
  def create
    raise "current_user does not have any division." if !current_user.default_division

    @partner = Partner.find params[:partner]
    @invoice = create_from_formdata(@partner, nil)
    @company = @invoice.company_setting
    journals, sum_amount = journals_from_formdata

    # 請求書パートに金額を合わせる
    d_amount = @invoice.si_details.inject(0) {|acc, x| acc + x.amount}
    journals.first.amount += d_amount + @invoice.ex_vat_amount - sum_amount

    journals.each {|j| @invoice.si_journals << j }
    SalesInvoice.transaction {
      if @invoice.save
        flash[:notice] = 'SalesInvoice was successfully created.'
        redirect_to :action => 'show', :id => @invoice
      else
        render :action => 'new'
      end
    } # of transaction
  end
  
  # 編集画面
  def edit
    @invoice = SalesInvoice.find(params[:id])
    @company = @invoice.company_setting
    @partner = @invoice.partner

    if !@invoice.editable_by(current_user)
      flash[:alert] = "この請求書は編集できません。"
      redirect_to :action => 'show', :id => @invoice
    end
  end
  
  # 自分が最後に更新したログ
  def last_modify_log
    il = SiLog.find :first,
           :conditions => ['sales_invoice_id = ? AND done_at = (select max(done_at) FROM si_logs WHERE sales_invoice_id = ?)', @invoice.id, @invoice.id]
    if il && il.kind == SiLog::MODIFY && il.done_by.id == current_user.id
      return il
    end

    return nil
  end
  private :last_modify_log

  # 更新する
  def update
    orig_invoice = SalesInvoice.find params[:id]
    @partner = orig_invoice.partner
    @company = orig_invoice.company_setting

    if !orig_invoice.editable_by(current_user)
      @invoice = orig_invoice
      flash[:alert] = "この請求書は編集できません。"
      redirect_to :action => 'show', :id => @invoice
    end

    begin
      SalesInvoice.transaction {
        @invoice = create_from_formdata @partner, orig_invoice
        
        # 印刷ずみのときのみ担当者変更
        @invoice.issued_by = current_user if @invoice.issued_by
        @invoice.save!
        
        # 更新日などを記録
        il = last_modify_log() || 
             SiLog.new(:sales_invoice => @invoice,
                       :kind => SiLog::MODIFY,
                       :done_by => current_user)
        il.attributes = {
          :remarks => (il.remarks || ""),
          :done_at => Time.now
        }
        il.save!
      }
    rescue ActiveRecord::RecordInvalid => e
      render :action => 'edit'
      return
    end

    flash[:notice] = '請求書は正常に更新されました。'
    redirect_to :action => 'show', :id => @invoice
  end

  # 請求書発行依頼の表示
  def print_request
    @invoice = SalesInvoice.find params[:id]
  end
  
  # 削除
  def destroy
    inv = SalesInvoice.find params[:id]
    partner = inv.partner
    SalesInvoice.transaction {
      cond = ['sales_invoice_id = ?', inv.id]
      SiDetail.delete_all cond
      SiBank.delete_all cond
      SiJournal.delete_all cond
      SiLog.delete_all cond
      inv.destroy
    }
    flash[:notice] = '請求書を削除しました。'
    redirect_to :controller => 'partners', :action => 'show', :id => partner
  end

  # 承認を要するリスト（表示）
  def sign_off_list
    @invoices = SalesInvoice.find :all,
                    :conditions => ['state < ?', SalesInvoice::SIGNED_OFF]
  end


  # 経理承認
  def sign_off
    @invoice = SalesInvoice.find params[:id]
    SalesInvoice.transaction {
      if @invoice.state < SalesInvoice::SIGNED_OFF
        il = SiLog.new(:sales_invoice => @invoice,
                       :kind => SiLog::SIGN_OFF,
                       :remarks => "",
                       :done_by => current_user,
                       :done_at => Time.now)
        il.save!
        @invoice.state = SalesInvoice::SIGNED_OFF
        @invoice.save!
      end
    }
    flash[:notice] = '請求書を承認しました。'
    redirect_to :action => 'show', :id => @invoice
  end
  
  # コピーして新しい請求書を起票
  def copy_new
    @invoice = SalesInvoice.find params[:id]
    @company = @invoice.company_setting
    @partner = @invoice.partner
    @invoice.attributes = {
      :sell_day => ((@invoice.sell_day + 1) >> 1) - 1,
      :due_day => ((@invoice.due_day + 1) >> 1) - 1,
      :si_number => nil,
    }
    
    render :action => 'new'
  end

  # 請求書をキャンセル
  def cancel_invoice
    inv = SalesInvoice.find params[:id]
    SalesInvoice.transaction {
      inv.state = 30
      inv.save!
      il = SiLog.new :sales_invoice => inv,
                     :kind => InvoiceLog::CANCEL,
                     :remarks => params[:message] || "",
                     :done_by => current_user,
                     :done_at => Time.now
      il.save!
    }
    flash[:notice] = '請求書をキャンセルしました。'
    redirect_to :controller => 'partners', :action => 'show', :id => inv.partner
  end

  # 宛て先を変更 (表示)
  def change_partner
    @invoice = SalesInvoice.find params[:id]
  end

  # 請求先を変更する2
  def change_partner2
    @invoice = SalesInvoice.find params[:id]

    @search = PartnerSearchBox.new(params[:search])
    @partners = @search.find_by_text
  end
  
  def change_partner_do
    @invoice = SalesInvoice.find params[:id]
    @partner = Partner.find params[:partner]
    SalesInvoice.transaction {
      @invoice.partner = @partner
      # @invoice.bank_account = nil
      il = SiLog.new(:sales_invoice => @invoice,
                     :kind => InvoiceLog::MODIFY,
                     :remarks => '取引先を変更。',
                     :done_by => current_user,
                     :done_at => Time.now)
      il.save!
      if @invoice.save
        flash[:notice] = '請求書は正常に更新されました。'
        redirect_to :action => 'show', :id => @invoice
      else
        raise
      end
    }
  end


  # 仕訳の編集（画面）
  def edit_journals
    @invoice = SalesInvoice.find params[:id]
    @company = @invoice.company_setting
  end


  # 仕訳の更新
  def update_journals
    @invoice = SalesInvoice.find params[:id]
    journals, sum_amount = journals_from_formdata

    # 請求書パートの金額と一致しているか
    d_amount = SiDetail.find :first,
                    :select => 'sum(amount) as sub_total',
                    :conditions => ['sales_invoice_id = ?', @invoice.id]
    raise if !d_amount || !d_amount.sub_total
    
    if sum_amount != d_amount.sub_total.to_i + @invoice.ex_vat_amount
      @invoice.errors.add_to_base('合計金額が請求金額と一致していません。')
      render :action => 'edit_journals'
      return
    end

    SalesInvoice.transaction {
      @invoice.si_journals.clear
      # SiJournal.delete_all ['sales_invoice_id = ?', @invoice.id]
      journals.each {|j| @invoice.si_journals << j }

      @invoice.save!

      il = last_modify_log() || 
           SiLog.new(:sales_invoice => @invoice,
                     :kind => SiLog::MODIFY,
                     :done_by => current_user)
      il.attributes = {
        :remarks => (il.remarks || "仕訳を更新。"),
        :done_at => Time.now
      }
      il.save!
    }

    flash[:notice] = '請求書は正常に更新されました。'
    redirect_to :action => 'show', :id => @invoice
  end
end
